; ideas:
; 1 master menu program - beep and blink selection when made
; 2 play a looping song, use keypad to change the key (use a table of half-step frequencies)
; 3 volume control? by controlling duty cycle of "on" frequency. Echoy sound, with a different pitch for each key
; 4 startup sound
; 5 some kind of real-time demo... 1-d wack a mole? -, 0, 8
;   maze slice - flash the current position
;   side scroller - flash current pos
;   pick up and put down items, move them side-to-side, cursor is ','
;   steer a car with scroll buttons? 
;   guess the number
;   chess - http://chessprogramming.wikispaces.com/Z80
;   battleship
;   hunt the wumpus/cave
;   artillery: firing angle, under/overshot by x feet
;   matrix-style name
;   random sound effect and display for each key



card_f2f_buffer = 0x8000    ; must start on a 256 byte boundary
random_seed = 0x8400
keypad_last_keycode = 0x8402
keypad_shift = 0x8403
keypad_buffer = 0x8404
card_f2f_buffer_full = 0x8405
card_f2f_buffer_end = 0x8406
card_f2f_time_constant = 0x8407
card_result_buffer = 0x8500
card_display_temp = 0x8408

start:
    DI
    JP main
  
    code @ 0x0018
    DW card_interrupt_handler
        
    code @ 0x0100
        
main:    
    LD SP,0xFFFF        ; init stack pointer
    IM 2                ; use mode 2 interrupts
    LD A,00H            ; interrupt vectors in page 0
    LD I,A
 	
    ; init port 0 - controls the display (bits 6,5,4) and keypad columns (bits 3,2,1,0)
    ; port 1 is the control register for port 0
    LD A,0xCF           ; we want to control each port bit individually
    OUT (0x01),A	
    LD A,0x80           ; bit 7 is input, others are outputs
    OUT (0x01),A
    LD A,0x18           ; use interrupt vector 18
    OUT (0x01),A
    LD A,0x97           ; generate interrupt if any masked bit is low
    OUT (0x01),A
    LD A,0x7F           ; mask = bit 7
    OUT (0x01),A
	
    LD A,0x3F           ; set the initial output values for port 0	
    OUT (0x00),A
	
    ; initialize the display
    IN A,(0x00)         ; get current port state
    AND 0xEF            ; clear bit 4 (display reset)
    OUT (0x00),A
    LD B,0x1C           ; wait 0x1C cycles
.L1 DJNZ .L1
    OR 0x10             ; set bit 4 (display reset)
    OUT (0x00),A 
    LD C,0xFF           ; set the display duty cycle to 31 (maximum brightness)
    CALL display_send_byte
    
    ; initalize the speaker
    LD A,0x03           ; disable interrupts, timer mode, prescale=16, auto-trigger, no time constant
    OUT (0x12),A
    
    ; initialize random seed	
    LD HL,0x4C0E        
    LD (random_seed),HL
    	
    ; startup sound	
    LD C,0xFE
    CALL speaker_beep_period
    CALL speaker_beep
    
    ; initialize the keypad input
    LD HL,keypad_last_keycode
    LD (HL),0x00
    LD HL,keypad_shift
    LD (HL),0x00
    
    ; initialize the card reader input
    LD A,0x00
    LD (card_f2f_buffer_full),A
    LD (card_f2f_buffer_end),A
    LD (card_f2f_buffer_end+1),A

    ; print a prompt message
    CALL display_clear
    LD HL,test_msg
    CALL display_print_string
    
    EI
    LD HL,card_f2f_buffer_full
     
wait_for_card:
    LD A,(HL)
    OR A
    JR Z,wait_for_card

got_card: 
    ; show card data    
    LD C,0xC0
    CALL speaker_beep_period
    
    ; clear flag
    LD DE,card_f2f_buffer_full
    LD A,0x00
    LD (DE),A
    
    ; parse the raw data
    LD HL,card_result_buffer
    LD DE,card_f2f_buffer+8
    LD A,(DE)
    SRL A                           ; a * 0.5
    ADD A,A                         ; a * 1.5
    SRL A                           ; a * 0.75
    LD (card_f2f_time_constant),A   ; zero/one bit threshold
    LD B,A                          

find_1_bit:    
    ; advance to the first one bit
    INC DE
    LD A,(DE)
    CP B
    JR NC,find_1_bit
    
next_bit:    
    LD A,(DE)
    CP B
    JR NC,.L1
    LD A,0x01
    INC DE
    INC DE    
    JR .L2
.L1 LD A,0x00
    INC DE
.L2 LD (HL),A
    INC HL
    LD A,D
    CP 0x84
    JR NC,build_string
    JR next_bit    

build_string:
    ; convert bits to characters
    LD HL,card_result_buffer
    LD DE,card_result_buffer
.L2 LD B,0x04
    LD C,0x00
.L1 LD A,(DE)
    INC DE
    RR A                           ; put A bit 0 in carry flag
    RR C                           ; shift carry flag in B bit 7
    DJNZ .L1
    
    ; check parity bit
    LD A,C
    OR A
    JR NZ,.L9                ; skip parity check for 0 data      
    INC DE
    JR .L8
    
.L9:    
    LD A,(DE)
    INC DE
    SLA A
    SLA A
    SLA A                           ; parity in bit 3
    OR C                            ; data in bits 7-4 
    SRL A
    SRL A
    SRL A                           ; data and parity in bits 4-0
    
    LD B,0x00                       ; parity count
    SRL A
    JR NC,.L3
    INC B
.L3 SRL A
    JR NC,.L4
    INC B
.L4 SRL A
    JR NC,.L5
    INC B
.L5 SRL A
    JR NC,.L6
    INC B
.L6 SRL A
    JR NC,.L7
    INC B
.L7 ; test parity
    LD A,B
    AND 0x01
    JP Z, parity_error
    
.L8:           
    ; process data
    SRL C
    SRL C
    SRL C
    SRL C                           
    LD A,C
    ADD A,'0'
    LD (HL),A
    INC HL
  
    CP '?'
    JR Z,string_done
    
    LD A,D
    CP 0x87
    JR NC,string_done    
    JR .L2
     
string_done:
    LD A,0x00
    LD (HL),A                   ; null terminate     
    LD HL,card_result_buffer
    JR scroll_next

parity_error:
    CALL display_clear
    LD HL,parity_error_msg
    CALL display_print_string
    LD HL,card_f2f_buffer_full
    JP wait_for_card
                
parity_error_msg:
    DB "YOUR CARD SMELLS",0
                    
new_character:
    LD C,0x00
.L1 LD A,(DE)
    CP B                            ; carry = 1 for 1, 0 for 0
    RR C
    INC DE
    JR .L1    
    
    CALL display_clear
    LD HL,test_msg_2                ; "END = "
    CALL display_print_string   
    
    LD DE,(card_f2f_buffer_end)
    CALL display_print_hex_word     ; show the buffer end address
    
    CALL keypad_get_next            ; wait for key
    
    LD HL,card_f2f_buffer
    
    CALL print_current

scroll:   
    CALL keypad_read
    OR A
    JR NZ,.L2
    LD DE,card_f2f_buffer_full
    LD A,(DE)
    OR A
    JP NZ,got_card                  ; new card     
    JR scroll
.L2 CP 0x0D                         ; scroll left?
    JR Z,scroll_left
    CP 0x0F                         ; scroll right?
    JR Z,scroll_right
    JR scroll
scroll_left:
    LD DE,0xFFF0
    ADD HL,DE
    JR scroll_next
scroll_right:
    LD DE,0x0010
    ADD HL,DE   
scroll_next: 
    CALL print_current  
    LD BC,0x0096
    CALL wait_ms
    JR scroll    
 
print_current:
    EX DE,HL                        ; string base ptr into DE
    LD HL,0x0010
    ADD HL,DE
    EX DE,HL
    LD A,(DE)                       ; save previous
    LD (card_display_temp),A
    LD A,0x00                       ; write null
    LD (DE),A
    CALL display_clear
    PUSH HL
    CALL display_print_string
    POP HL
    
    EX DE,HL                        ; string base ptr into DE
    LD HL,0x0010
    ADD HL,DE
    EX DE,HL
    LD A,(card_display_temp)
    LD (DE),A                       ; restore original value of nulled byte
    
    RET
           
chat:    
    ; print a prompt message
    CALL display_clear
    LD HL,test_msg
    CALL display_print_string
    
    ; get response
    CALL keypad_get_string
    
    ; print result
    CALL display_clear
    LD HL,test_msg_2
    CALL display_print_string
    LD HL,keypad_buffer
    CALL display_print_string
    
    LD BC,0xFA0         ; wait 4000ms
    CALL wait_ms
    CALL speaker_beep
    JR chat   
    
    
    
echo:    
    CALL keypad_get_next
    
    PUSH AF
    CALL speaker_beep
    POP AF
    
    CALL keycode_to_ascii 
    LD C,A
    CALL display_send_byte
    
    JR echo
    
    
    
    
    ; --- play noise    
.L4 LD A,0x05           ; time constant follows
    OUT (0x12),A
    CALL random_get_next              
    OUT (0x12),A        ; speaker on
    JR .L4
    	
    		
    ; --- siren
    LD D,0xFF
    
.L2 LD A,0x05           ; time constant follows
    OUT (0x12),A
    LD A,D              
    OUT (0x12),A        ; speaker on
    LD BC,0x0010	   ; wait 16 ms
    CALL wait_ms
    DEC D
    LD A,0x40
    CP D
    JR NZ,.L2
   
.L3 LD A,0x05           ; time constant follows
    OUT (0x12),A
    LD A,D
    OUT (0x12),A        ; speaker on
    LD BC,0x0010	   ; wait 16 ms
    CALL wait_ms
    INC D
    JR NZ,.L3
    
    DEC D    
    JR .L2



test_msg:
    DB "SWIPE CARD ",0
test_msg_2:
    DB "END = ",0
      
display_print_string:
    ; pointer to the string is in HL
    ; the string must be terminated with a 0 byte
    LD A,(HL)
    INC HL
    OR A
    RET Z
    LD C,A
    CALL display_send_byte
    JR display_print_string

display_clear:
    LD C,0xAF           ; set the cursor to the beginning of the line
    CALL display_send_byte
    LD D,0x10           ; 16 spaces to print
.L1 LD C,' '
    CALL display_send_byte
    DEC D
    JR NZ,.L1  
    RET
    
display_print_hex_word:
    ; word to be printed is in DE
    PUSH DE
    LD C,D
    CALL display_print_hex_byte
    POP DE
    LD C,E
    JR display_print_hex_byte
         
display_print_hex_byte:
    ; byte to be printed is in C
    PUSH BC
    LD A,C
    SRL A
    SRL A
    SRL A
    SRL A
    CALL display_print_hex_nibble
    POP BC
    LD A,C
    AND 0x0F
    JR display_print_hex_nibble
    
display_print_hex_nibble:
    ; nibble to be printed is in A3-A0
    ADD A,'0'           ; add '0'
    CP 0x3A
    JR C,.L1            ; result less than ':'?
    ADD A,0x07          ; make relative to 'A'
.L1 LD C,A
    JR display_send_byte
            	
display_send_byte:
    ; byte to be sent is in C
    ; note the display controller does not support lower-case letters!
    LD B,0x08           ; 8 bits to send
.L1 IN A,(0x00)         ; get current port state
    RLA                 ; rotate the port word until the data bit is in the carry flag
    RLA
    RLA
    RL C                ; shift the next output data bit into the carry flag
    RRA                 ; rotate the port word until the data bit is in bit 5
    RRA
    RRA
    OUT (0x00),A        ; setup the output bit
    OR 0x40             ; set clock high (bit 6)
    OUT (0x00),A
    AND 0xBF            ; set clock low (bit 6)
    OUT (0x00),A
    DJNZ .L1            ; continue with the next bit
    RET
	
speaker_beep:	
    LD C,0x7F           ; default time const = 128*16 clocks = 512us @ 4MHz clock = 1953Hz
speaker_beep_period:
    ; beep tone period time constant should be passed in C
    LD A,0x07           ; disable interrupts, timer mode, prescale=16, auto-trigger, time constant follows
    OUT (0x12),A
    LD A,C              ; time const
    OUT (0x12),A        ; speaker on
    LD BC,0x0096	   ; wait 150 ms
    CALL wait_ms
    LD A,0x03
    OUT (0x12),A        ; speaker off
    RET

keypad_get_string:
    LD HL,keypad_shift
    LD (HL),0x00  
    LD HL,keypad_buffer
    LD (HL),0x00        ; initialize input buffer to empty string
    ; --- KEYPRESS loop
.L1 CALL keypad_get_next
    CP 0x04            ; CLEAR key?
    JR Z,.L4           ; ignore it
    CP 0x08            ; BACKSPACE key?
    JR Z,.L2
    CP 0x0C            ; ALPHA key?
    JR Z,.L3
    CP 0x10            ; ENTER key?
    JR Z,.L8  
    ; --- NORMAL KEYPRESS handler
    PUSH HL
    LD HL,keypad_last_keycode
    LD (HL),A           ; remember the last regular keycode
    LD HL,keypad_shift
    LD (HL),0x00        ; reset shift
.L7 CALL keycode_to_ascii
    POP HL
    LD (HL),A           ; append to string
    INC HL
    LD (HL),0x00        ; add zero terminator
    JR .L4     
    ; --- BACKSPACE handler
.L2 LD A,H              
    CP high keypad_buffer
    JR NZ,.L5
    LD A,L
    CP low keypad_buffer    
    JR Z,.L4            ; ignore backspace if at start of string           
.L5 DEC HL              ; perform backspace
    LD (HL),0x00        ; add zero terminator
    PUSH HL
    LD HL,keypad_shift
    LD (HL),0x00        ; reset shift
    POP HL
    JR .L4    
    ; --- ALPHA handler
.L3 LD A,H              
    CP high keypad_buffer
    JR NZ,.L6
    LD A,L
    CP low keypad_buffer    
    JR Z,.L4            ; ignore alpha if at start of string  
.L6 DEC HL              ; perform alpha
    PUSH HL             
    LD HL,keypad_shift  
    LD A,(HL)
    INC A               ; get new shift value
    AND 0x03  
    LD (HL),A
    LD HL,keypad_last_keycode
    LD A,(HL)
    JR .L7              ; replay the last keycode with the new shift value
    ; --- ENTER handler
.L8 LD HL,keypad_shift
    LD (HL),0x00        ; reset shift
    LD BC,0x00FF	   ; wait 255 ms
    CALL wait_ms      
    RET     
    ; --- POST KEYPRESS handler    
.L4 PUSH HL
    LD HL,keypad_buffer
    CALL display_clear
    CALL display_print_string
    POP HL
    LD BC,0x00FF	   ; wait 255 ms
    CALL wait_ms    
    JR .L1
        
keypad_get_next:
    CALL keypad_read
    OR A
    JR Z, keypad_get_next
    RET
    
keypad_read:
    ; columns can be activated by setting port 0 bits 3-0 low
    ; active rows be detected by sensing port 2 bits 3-0 low
    LD C,0xFE           ; initial column mask
.L1 IN A,(0x00)         ; get current port state
    OR 0x0F             ; disable all columns
    AND C               ; activate next column
    OUT (0x00),A        
    IN A,(0x02)         ; read result
    AND 0x0F            ; mask the rows
    CP 0x0F             ; any rows active?
    JR NZ,keypad_read_key
    RLC C               ; adjust column mask
    LD A,0xEF           ; was that the last column?
    CP C
    JR NZ,.L1           ; next column
    LD A,0x00           ; no keys are pressed
    RET
keypad_read_key:    
    LD B,A              ; put row data in B
    LD A,0x00           ; initialize key code
.L4 SRA B               ; shift right row data
    JR NC,.L3           ; was that the active row?
    ADD A,0x04          ; no, add 4 to key code
    JR .L4
.L3 SRA C               ; shift right column data
    JR NC,.L5           ; was that the active column?
    INC A               ; no, add 1 to key code
    JR .L3
.L5 INC A               ; convert to 1-based key code
    RET

keycode_to_ascii:
    ; keycode is in A, result is returned in A
    SLA A               ; multiply keycode by 4
    SLA A
    LD HL,keypad_shift
    ADD A,(HL)          ; add the shift value
    LD E,A
    LD D,0x00
    LD HL,keycode_table-4
    ADD HL,DE
    LD A,(HL)
    RET
    
keycode_table:
    DB "1QZ.2ABC3DEFA   4GHI5JKL6MNOB   7PRS8TUV9WXYC   *,'\"0- +#:;@D   "

card_interrupt_handler:
    EXX
    EX AF,AF'
    CALL card_read
    LD (card_f2f_buffer_end),HL
    LD HL,card_f2f_buffer_full
    LD A,0x01
    LD (HL),A
    CALL speaker_beep
    EX AF,AF'
    EXX
    EI
    RETI

card_read:  
    ; read card F2F transitions into card_f2f_buffer
    LD HL,card_f2f_buffer
    LD DE,card_f2f_buffer+0x401
    LD BC,0100H
    CALL card_wait_for_high
    RET
    
card_low:  
    INC HL
    LD A,H
    CP D
    RET NC
    LD B,E
card_wait_for_high:
    IN A,(C)
    JP M,card_high
    INC B
    LD (HL),B
    JR NZ,card_wait_for_high
    RET
card_wait_for_low:
    IN A,(C)
    JP P,card_low
card_high:
    INC B
    LD (HL),B
    JR NZ,card_wait_for_low
    RET
            
wait_ms:
    ; delays BC milliseconds
    PUSH BC 
    LD BC,0086H
.L1 DEC BC
    LD A,B
    OR C
    JR NZ,.L1
    POP BC
    DEC BC
    LD A,B
    OR C
    JR NZ,wait_ms
    RET
    
random_get_next:
    PUSH HL
    PUSH DE
    LD HL,(random_seed)
    LD A,R
    LD D,A
    LD E,(HL)
    ADD HL,DE
    ADD A,L
    XOR H
    LD (random_seed),HL
    POP DE
    POP HL
    RET    